//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <fstream.h>
#pragma hdrstop

#include "transform.h"
#include "DCTvsDWT.h"
#include "About.h"
#include "Threads.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TMainWindow *MainWindow;
//---------------------------------------------------------------------------
__fastcall TMainWindow::TMainWindow(TComponent* Owner)
	: TForm(Owner)
{
    trGroup=new TransformGroup[6];

    trGroup[0].tabSheet=FourierTabSheet;
    trGroup[0].image=FourierImage;
	trGroup[0].showGrid=FourierGridBox;
    trGroup[0].correction=FourierColorCorrection;
	trGroup[0].errorBar=FourierErrorTrackBar;
	trGroup[0].errorLabel=FourierErrorLabel;
	trGroup[0].gridBar=FourierWinSizeTrackBar;
	trGroup[0].gridLabel=FourierWinSizeLabel;
    trGroup[0].compressBar=FourierCompressBar;
	trGroup[0].compressLabel=FourierCompressLabel;
    trGroup[0].transform=new CosineFFT;
    trGroup[0].zig=Transformer::ZigZag;

    trGroup[3].tabSheet=Fourier2DTabSheet;
    trGroup[3].image=Fourier2DImage;
	trGroup[3].showGrid=Fourier2DGridBox;
    trGroup[3].correction=Fourier2DColorCorrection;
	trGroup[3].errorBar=Fourier2DErrorTrackBar;
	trGroup[3].errorLabel=Fourier2DErrorLabel;
	trGroup[3].gridBar=Fourier2DWinSizeTrackBar;
	trGroup[3].gridLabel=Fourier2DWinSizeLabel;
    trGroup[3].compressBar=Fourier2DCompressBar;
	trGroup[3].compressLabel=Fourier2DCompressLabel;
    trGroup[3].transform=new CosineFFT2D;
    trGroup[3].zig=Transformer::Normal;

    trGroup[2].tabSheet=HaarTabSheet;
    trGroup[2].image=HaarImage;
	trGroup[2].showGrid=HaarGridBox;
    trGroup[2].correction=HaarColorCorrection;
	trGroup[2].errorBar=HaarErrorTrackBar;
	trGroup[2].errorLabel=HaarErrorLabel;
	trGroup[2].gridBar=HaarWinSizeTrackBar;
	trGroup[2].gridLabel=HaarWinSizeLabel;
    trGroup[2].compressBar=HaarCompressBar;
	trGroup[2].compressLabel=HaarCompressLabel;
    trGroup[2].transform=new HaarDWT;
    trGroup[2].zig=Transformer::ZigZag;

    trGroup[5].tabSheet=Haar2DTabSheet;
    trGroup[5].image=Haar2DImage;
	trGroup[5].showGrid=Haar2DGridBox;
    trGroup[5].correction=Haar2DColorCorrection;
	trGroup[5].errorBar=Haar2DErrorTrackBar;
	trGroup[5].errorLabel=Haar2DErrorLabel;
	trGroup[5].gridBar=Haar2DWinSizeTrackBar;
	trGroup[5].gridLabel=Haar2DWinSizeLabel;
    trGroup[5].compressBar=Haar2DCompressBar;
	trGroup[5].compressLabel=Haar2DCompressLabel;
    trGroup[5].transform=new HaarDWT2D;
    trGroup[5].zig=Transformer::Normal;

    trGroup[1].tabSheet=DaubTabSheet;
    trGroup[1].image=DaubImage;
	trGroup[1].showGrid=DaubGridBox;
    trGroup[1].correction=DaubColorCorrection;
	trGroup[1].errorBar=DaubErrorTrackBar;
	trGroup[1].errorLabel=DaubErrorLabel;
	trGroup[1].gridBar=DaubWinSizeTrackBar;
	trGroup[1].gridLabel=DaubWinSizeLabel;
    trGroup[1].compressBar=DaubCompressBar;
	trGroup[1].compressLabel=DaubCompressLabel;
    trGroup[1].transform=new Daubechies4DWT;
    trGroup[1].zig=Transformer::ZigZag;

    trGroup[4].tabSheet=Daub2DTabSheet;
    trGroup[4].image=Daub2DImage;
	trGroup[4].showGrid=Daub2DGridBox;
    trGroup[4].correction=Daub2DColorCorrection;
	trGroup[4].errorBar=Daub2DErrorTrackBar;
	trGroup[4].errorLabel=Daub2DErrorLabel;
	trGroup[4].gridBar=Daub2DWinSizeTrackBar;
	trGroup[4].gridLabel=Daub2DWinSizeLabel;
    trGroup[4].compressBar=Daub2DCompressBar;
	trGroup[4].compressLabel=Daub2DCompressLabel;
    trGroup[4].transform=new Daubechies4DWT2D;
    trGroup[4].zig=Transformer::Normal;

    for(int i=0;i<6;i++)
    {
        trGroup[i].dirtyOriginal=false;
	    trGroup[i].dirty=false;
    	trGroup[i].previousGridBar=trGroup[i].gridBar->Position;
    }
	int activePage=TransformPageControl->ActivePage->PageIndex;
    activeGroup=&trGroup[activePage];
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::SelectImageToView(TObject *Sender)
{
	QuickViewImage->Picture->Bitmap->LoadFromFile(dynamic_cast<TFileListBox *>(Sender)->FileName);
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::ExitButtonClick(TObject *Sender)
{
    for(int i=0;i<6;i++)
    {
        if(trGroup[i].dirty)
        {
        	AnsiString pageCaption=AnsiString(trGroup[i].tabSheet->Caption);
        	AnsiString text=pageCaption+AnsiString(" transformed image not saved. Do you want to save ?");
			switch(Application->MessageBox(text.c_str(),"Be careful !",MB_YESNOCANCEL))
            {
            	case IDOK:		SaveImageDialog->Title=AnsiString("Save image : ")+pageCaption;
                                if(SaveImageDialog->Execute())
                                {
                                    imagePath=SaveImageDialog->FileName;
                                    trGroup[i].image->Picture->SaveToFile(imagePath);
                                    SaveButton->Enabled=false;
                                    trGroup[i].dirty=false;
                                }
                break;
                case IDCANCEL:  return;
                default:		break;
            }
        }
    }
	Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::OpenButtonClick(TObject *Sender)
{
	if(OpenImageDialog->Execute())
    {
    	imagePath=OpenImageDialog->FileName;
    	OriginalImage->Picture->Bitmap->LoadFromFile(imagePath);
	    OriginalTabSheet->Caption=imagePath;
    	OriginalPageControl->ActivePage=OriginalTabSheet;
        ConvertButton->Enabled=true;
    }
    for(int i=0;i<6;i++)
    {
        trGroup[i].dirtyOriginal=true;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::SaveButtonClick(TObject *Sender)
{
	if(SaveImageDialog->Execute())
    {
    	imagePath=SaveImageDialog->FileName;
       	activeGroup->image->Picture->SaveToFile(imagePath);
		SaveButton->Enabled=false;
	    activeGroup->dirty=false;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::SelectImageToConvert(TObject *Sender)
{
	imagePath=dynamic_cast<TFileListBox *>(Sender)->FileName;
   	OriginalImage->Picture->Bitmap->LoadFromFile(imagePath);
	QuickViewImage->Picture->Bitmap->LoadFromFile(imagePath);
    OriginalTabSheet->Caption=imagePath;
    OriginalPageControl->ActivePage=OriginalTabSheet;
    ConvertButton->Enabled=true;

    for(int i=0;i<6;i++)
    {
        trGroup[i].dirtyOriginal=true;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::ConvertButtonClick(TObject *Sender)
{
    int 	width;
    int 	grid;
    int 	color,colorRed,colorGreen,colorBlue;
    double 	compress,redCompress,greenCompress,blueCompress;
    double 	error=static_cast<double>(activeGroup->errorBar->Position)/100.0;
	char 	buffer[10];
    char	tpl[]="%3.2f %";

	MainWindow->Cursor=crHourGlass;

    activeGroup->image->Picture->Bitmap=OriginalImage->Picture->Bitmap;
    grid=1<<(activeGroup->gridBar->Position-1);
	width=activeGroup->image->Width;

    Transformer transRed(width,grid,error,activeGroup->transform);
    Transformer transGreen(width,grid,error,activeGroup->transform);
    Transformer transBlue(width,grid,error,activeGroup->transform);

    TCanvas *canvas=activeGroup->image->Picture->Bitmap->Canvas;
    for(int i=0;i<width;i++)
    {
        for(int j=0;j<width;j++)
        {
            color=::GetPixel(canvas->Handle,i,j);
            transRed[i*width+j]=color & clRed;
            transGreen[i*width+j]=(color & clLime)>>8;
            transBlue[i*width+j]=(color & clBlue)>>16;
        }
        TransformProgressBar->Position=static_cast<TProgressRange>(i);
    }
    TransformProgressBar->Position=0;

    transRed.Process(activeGroup->zig);
    redCompress=transRed.Compress();
    sprintf(buffer,tpl,redCompress);
    activeGroup->compressLabel->Caption=AnsiString("Red channel: ")+AnsiString(buffer);
    activeGroup->compressBar->Position=redCompress;
    TransformProgressBar->Position=84;

    transGreen.Process(activeGroup->zig);
    greenCompress=transGreen.Compress();
    sprintf(buffer,tpl,greenCompress);
    activeGroup->compressLabel->Caption=AnsiString("Green channel: ")+AnsiString(buffer);
    activeGroup->compressBar->Position=greenCompress;
    TransformProgressBar->Position=168;

    transBlue.Process(activeGroup->zig);
    blueCompress=transBlue.Compress();
    sprintf(buffer,tpl,blueCompress);
    activeGroup->compressLabel->Caption=AnsiString("Blue channel: ")+AnsiString(buffer);
    activeGroup->compressBar->Position=blueCompress;
    TransformProgressBar->Position=256;

    if(activeGroup->correction->Checked)
    {
    	ScaleTo<int,255> scaleRed(transRed.Bitmap(),width*width);
    	ScaleTo<int,255> scaleGreen(transGreen.Bitmap(),width*width);
    	ScaleTo<int,255> scaleBlue(transBlue.Bitmap(),width*width);
    }
    canvas->Pixels[0][0]=static_cast<TColor>(0x000000);
    for(int i=0;i<width;i++)
    {
        for(int j=0;j<width;j++)
        {
        	colorRed=RoundTo<int,255>(transRed[i*width+j]) & clRed;
            colorGreen=(RoundTo<int,255>(transGreen[i*width+j])<<8) & clLime;
            colorBlue=(RoundTo<int,255>(transBlue[i*width+j])<<16) & clBlue;
        	color=(colorRed | colorGreen | colorBlue) & clWhite;

            ::SetPixel(canvas->Handle,i,j,ColorToRGB(static_cast<TColor>(color)));
        }
        TransformProgressBar->Position=static_cast<TProgressRange>(i);
    }
    TransformProgressBar->Position=0;

	SaveButton->Enabled=true;
    activeGroup->dirty=true;

	MainWindow->Cursor=crArrow;
    compress=(redCompress+greenCompress+blueCompress)/3.0;
    sprintf(buffer,tpl,compress);
    activeGroup->compressLabel->Caption=AnsiString("Compress: ")+AnsiString(buffer);
    activeGroup->compressBar->Position=compress;
    activeGroup->dirtyOriginal=false;

	if(activeGroup->showGrid->Checked)
    {
        TCanvas *canvas=activeGroup->image->Canvas;
        canvas->Pen->Mode=pmNot;
        for(int i=width/grid;i<width;i+=(width/grid))
        {
            canvas->MoveTo(0,i);
            canvas->LineTo(width,i);
            canvas->MoveTo(i,width);
            canvas->LineTo(i,0);
        }
        canvas->Pen->Mode=pmCopy;
    }
	TransformThread *redThread=new TransformThread(false,activeGroup,ConvertButton);
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::AboutButtonClick(TObject *Sender)
{
	AboutDialog->ShowModal();
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::ChangeErrorLabel(TObject *Sender)
{
	activeGroup->errorLabel->Caption=AnsiString("Error: ")+AnsiString(activeGroup->errorBar->Position)+AnsiString("%");
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::ZoomInImage(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
	TImage *image=activeGroup->image;
    if(Button==mbLeft)
    {
    	image->Width=image->Width*2;
    	image->Height=image->Height*2;
    }
    if(Button==mbRight)
    {
    	image->Width=image->Width/2;
    	image->Height=image->Height/2;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::ChangeActivePage(TObject *Sender)
{
	int activePage=dynamic_cast<TPageControl *>(Sender)->ActivePage->PageIndex;
    activeGroup=&trGroup[activePage];
	SaveButton->Enabled=activeGroup->dirty;
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::DrawGrid(TObject *Sender)
{
   	if(activeGroup->dirtyOriginal)
    {
    	activeGroup->image->Picture->Bitmap->LoadFromFile(imagePath);
        activeGroup->dirtyOriginal=false;
    }
   	int oldGrid=1<<(activeGroup->previousGridBar-1);
    int grid=1<<(activeGroup->gridBar->Position-1);
    int height=activeGroup->image->Height;
    int width=activeGroup->image->Width;

	if(activeGroup->showGrid->Checked && oldGrid!=grid)
    {
        TCanvas *canvas=activeGroup->image->Canvas;
	    canvas->Pen->Mode=pmNot;
		//...	clear previous one
        for(int i=width/oldGrid;i<width;i+=(width/oldGrid))
        {
            canvas->MoveTo(i,height);
            canvas->LineTo(i,0);
            canvas->MoveTo(0,i);
            canvas->LineTo(width,i);
        }
        //...	draw new one
        for(int i=width/grid;i<width;i+=(width/grid))
        {
            canvas->MoveTo(0,i);
            canvas->LineTo(width,i);
            canvas->MoveTo(i,height);
            canvas->LineTo(i,0);
        }
        canvas->Pen->Mode=pmCopy;
	    activeGroup->previousGridBar=activeGroup->gridBar->Position;
    }
    activeGroup->gridLabel->Caption=AnsiString("Grid size: ")+AnsiString(width/grid);
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::ZoomOriginalImage(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
    if(Button==mbLeft)
    {
    	OriginalImage->Width=OriginalImage->Width*2;
    	OriginalImage->Height=OriginalImage->Height*2;
    }
    if(Button==mbRight)
    {
    	OriginalImage->Width=OriginalImage->Width/2;
    	OriginalImage->Height=OriginalImage->Height/2;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainWindow::ShowGrid(TObject *Sender)
{
    int grid=1<<(activeGroup->gridBar->Position-1);
    int width=activeGroup->image->Width;
    {
        TCanvas *canvas=activeGroup->image->Canvas;
    	canvas->Pen->Mode=pmNot;
        for(int i=width/grid;i<width;i+=(width/grid))
        {
            canvas->MoveTo(0,i);
            canvas->LineTo(width,i);
            canvas->MoveTo(i,width);
            canvas->LineTo(i,0);
        }
    	canvas->Pen->Mode=pmCopy;
	    activeGroup->previousGridBar=activeGroup->gridBar->Position;
    }
}
//--------------------------------------------------------------------------- 
