//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <fstream.h>
#pragma hdrstop

#include "transform.h"
#include "DCTvsDWT.h"
#include "Threads.h"
//---------------------------------------------------------------------------
//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall TransformThread::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------
__fastcall TransformThread::TransformThread(bool CreateSuspended,TransformGroup *tg,TSpeedButton *cb,TSpeedButton *sb,TProgressBar *pb)
						   :TThread(CreateSuspended),
                            transformGroup(tg),
                            convertButton(cb),
                            SaveButton(sb),
                            TransformProgressBar(pb)
{
}

__fastcall TransformThread::~TransformThread()
{
    delete transRed;
    delete transGreen;
    delete transBlue;
	return;
}
//---------------------------------------------------------------------------
void __fastcall TransformThread::Execute()
{

	//---- Place thread code here ----
    Synchronize(Arguments);

    transRed=new Transformer(width,grid,error,transformGroup->transform);
    transGreen=new Transformer(width,grid,error,transformGroup->transform);
    transBlue=new Transformer(width,grid,error,transformGroup->transform);

    Synchronize(CopyFromTables);

    transRed->Process(transformGroup->zig);
    redCompress=compress=transRed->Compress();
    Synchronize(Update);

    transGreen->Process(transformGroup->zig);
    greenCompress=compress=transGreen->Compress();
    Synchronize(Update);

    transBlue->Process(transformGroup->zig);
    blueCompress=compress=transBlue->Compress();
    Synchronize(Update);

    if(correction)
    {
    	ScaleTo<int,255> scaleRed(transRed->Bitmap(),width*width);
    	ScaleTo<int,255> scaleGreen(transGreen->Bitmap(),width*width);
    	ScaleTo<int,255> scaleBlue(transBlue->Bitmap(),width*width);
    }

    compress=(redCompress+greenCompress+blueCompress)/3.0;
    Synchronize(Update);

    Synchronize(CopyToImage);
}
//---------------------------------------------------------------------------

void __fastcall TransformThread::Arguments()
{
	transformGroup->busy=true;
	transformGroup->tabSheet->Cursor=crHourGlass;
    convertButton->Enabled=false;
	SaveButton->Enabled=false;

    error=static_cast<double>(transformGroup->errorBar->Position)/100.0;
    canvas=transformGroup->image->Picture->Bitmap->Canvas;
    grid=1<<(transformGroup->gridBar->Position-1);
	width=transformGroup->image->Width;

    correction=transformGroup->correction->Checked;
    grided=transformGroup->showGrid->Checked;
}

void __fastcall TransformThread::CopyFromTables()
{
    for(int i=0;i<width;i++)
    {
        for(int j=0;j<width;j++)
        {
            color=::GetPixel(canvas->Handle,i,j);
            (*transRed)[i*width+j]=color & clRed;
            (*transGreen)[i*width+j]=(color & clLime)>>8;
            (*transBlue)[i*width+j]=(color & clBlue)>>16;
        }
        TransformProgressBar->Position=static_cast<TProgressRange>(i);
    }
    TransformProgressBar->Position=0;
}

void __fastcall TransformThread::CopyToImage()
{
    TransformProgressBar->Position=0;
    canvas->Pixels[0][0]=static_cast<TColor>(0x000000);
    for(int i=0;i<width;i++)
    {
        for(int j=0;j<width;j++)
        {
        	colorRed=RoundTo<int,255>((*transRed)[i*width+j]) & clRed;
            colorGreen=(RoundTo<int,255>((*transGreen)[i*width+j])<<8) & clLime;
            colorBlue=(RoundTo<int,255>((*transBlue)[i*width+j])<<16) & clBlue;
        	color=(colorRed | colorGreen | colorBlue) & clWhite;

            ::SetPixel(canvas->Handle,i,j,ColorToRGB(static_cast<TColor>(color)));
        }
        TransformProgressBar->Position=static_cast<TProgressRange>(i);
    }
	if(grided)
    {
        TCanvas *canvas=transformGroup->image->Canvas;
        canvas->Pen->Mode=pmNot;
        for(int i=width/grid;i<width;i+=(width/grid))
        {
            canvas->MoveTo(0,i);
            canvas->LineTo(width,i);
            canvas->MoveTo(i,width);
            canvas->LineTo(i,0);
        }
        canvas->Pen->Mode=pmCopy;
    }
	SaveButton->Enabled=true;
    convertButton->Enabled=true;
	transformGroup->tabSheet->Cursor=crArrow;
	transformGroup->busy=false;
    transformGroup->dirty=true;
}

void __fastcall TransformThread::Update()
{
	char 	buffer[10];
    char	tpl[]="%3.2f %";

    sprintf(buffer,tpl,compress);
    transformGroup->compressLabel->Caption=AnsiString("Comp : ")+AnsiString(buffer);
    transformGroup->compressBar->Position=compress;
    TransformProgressBar->Position=TransformProgressBar->Position+84;
}
