//---------------------------------------------------------------------------
#ifndef transforH
#define transforH

typedef unsigned int uint;

template <class T>
inline
void Swap(T &a,T &b)
{
    T tempr=a;
    a=b;
    b=tempr;
}

inline
ostream &tab(ostream &os)
{
    return os<<"\t";
}

class Transform
{
public:
    enum Direction
    {
        Forward=1,
        Inverse=-1
    };
protected:
    double 	*ptr;
    double  compress;
    void 	realFFT(double *data,const unsigned int n,const Direction);
    void 	FFT(double *data,const unsigned int n,const Direction);
public:
	Transform():ptr(0),compress(0)		{	}
    virtual void    ForwardTransform(double *,const uint)=0;
    virtual void    InverseTransform(double *,const uint)=0;
    virtual void  	SortAndDiscard(double *,const uint,const double);
    virtual void    ReportTransform(const double *,const uint,ofstream &);
    double 			Compress() const	{	return compress;	}

};

class CosineFFT : public Transform
{
    void CosFFT(double *,const uint,const Direction);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class Daubechies4DWT : public Transform
{
    void Daub4DWT(double *,const uint,const Direction);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class HaarDWT : public Transform
{
    double *wavelet;

    void ForwardStep(const uint);
    void InverseStep(const uint);
    void _ForwardTransform(const uint);
    void _InverseTransform(const uint);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

//... inlines
inline
void CosineFFT::ForwardTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    CosFFT(ptr,nn,Forward);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
}

inline
void CosineFFT::InverseTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    CosFFT(ptr,nn,Inverse);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
};

inline
void Daubechies4DWT::ForwardTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    Daub4DWT(ptr,nn,Forward);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
}

inline
void Daubechies4DWT::InverseTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    Daub4DWT(ptr,nn,Inverse);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
}

inline
void HaarDWT::ForwardTransform(double *d,const uint nn)
{
    wavelet=new double[nn];
    ptr=d;
    _ForwardTransform(nn);
    delete []wavelet;
}

inline
void HaarDWT::InverseTransform(double *d,const uint nn)
{
    wavelet=new double[nn];
    ptr=d;
    _InverseTransform(nn);
    delete []wavelet;
}

inline
int QsortComparator(const void *p1, const void *p2)
{
	return (*static_cast<const double *>(p1) < *static_cast<const double *>(p2)) - (*static_cast<const double *>(p1) > *static_cast<const double *>(p2));
}
#endif
