#include <except.h>
#include <cstring.h>
#include <math.h>
#include <stdlib.h>
#include <fstream.h>
#include <vcl\vcl.h>

#include "transform.h"


const double C0=0.4829629131445341;
const double C1=0.8365163037378079;
const double C2=0.2241438680420134;
const double C3=-0.1294095225512604;

//... transform implementation
void Transform::FFT(double *data,const unsigned int nn,const Direction isign)
{
	unsigned int n,mmax,m,j,istep,i;
	double wtemp,wr,wpr,wpi,wi,theta;
	double tempr,tempi;

	n=nn << 1;
	j=1;
	for (i=1;i<n;i+=2)
    {
		if (j > i)
        {
			Swap(data[j],data[i]);
			Swap(data[j+1],data[i+1]);
		}
		m=n >> 1;
		while (m >= 2 && j > m)
        {
			j -= m;
			m >>= 1;
		}
		j += m;
	}
	mmax=2;
	while (n > mmax)
    {
		istep=mmax << 1;
		theta=isign*(2*M_PI/mmax);
		wtemp=sin(0.5*theta);
		wpr = -2.0*wtemp*wtemp;
		wpi=sin(theta);
		wr=1.0;
		wi=0.0;
		for (m=1;m<mmax;m+=2)
        {
			for (i=m;i<=n;i+=istep)
            {
				j=i+mmax;
				tempr=wr*data[j]-wi*data[j+1];
				tempi=wr*data[j+1]+wi*data[j];
				data[j]=data[i]-tempr;
				data[j+1]=data[i+1]-tempi;
				data[i] += tempr;
				data[i+1] += tempi;
			}
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
		}
		mmax=istep;
	}
}

void Transform::realFFT(double *data,const unsigned int n,const Direction isign)
{
	unsigned int i,i1,i2,i3,i4,np3;
	double c1=0.5,c2,h1r,h1i,h2r,h2i;
	double wr,wi,wpr,wpi,wtemp,theta;

	theta=M_PI/static_cast<double>(n>>1);
	if (isign==Forward)
    {
		c2 = -0.5;
		FFT(data,n>>1,isign);
	}
    else
    {
		c2=0.5;
		theta = -theta;
	}
	wtemp=sin(0.5*theta);
	wpr = -2.0*wtemp*wtemp;
	wpi=sin(theta);
	wr=1.0+wpr;
	wi=wpi;
	np3=n+3;
	for(i=2;i<=(n>>2);i++)
    {
		i4=1+(i3=np3-(i2=1+(i1=i+i-1)));
		h1r=c1*(data[i1]+data[i3]);
		h1i=c1*(data[i2]-data[i4]);
		h2r = -c2*(data[i2]+data[i4]);
		h2i=c2*(data[i1]-data[i3]);
		data[i1]=h1r+wr*h2r-wi*h2i;
		data[i2]=h1i+wr*h2i+wi*h2r;
		data[i3]=h1r-wr*h2r+wi*h2i;
		data[i4] = -h1i+wr*h2i+wi*h2r;
		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	if(isign==Forward)
    {
		data[1] = (h1r=data[1])+data[2];
		data[2] = h1r-data[2];
	}
    else
    {
		data[1]=c1*((h1r=data[1])+data[2]);
		data[2]=c1*(h1r-data[2]);
		FFT(data,n>>1,isign);
	}
}

void Transform::ReportTransform(const double *table,const uint nn,ofstream &os)
{
   	for(uint i=0;i<nn;i++)
    {
   		os<<table[i]<<tab;
    }
}

void Transform::SortAndDiscard(double *table,const uint s,const double error)
{
    uint 	size=s*s;
	double 	measure=0;
	double 	discardSum=0;

	double 	threshold;
	double 	approx;
	double  *sortedTable=new double[size];
    for(uint j=0;j<size;j++)
    {
        sortedTable[j]=fabs(table[j]);
        measure+=sortedTable[j]*sortedTable[j];
    }
    if(measure!=0.0)
    {
	    measure=sqrt(measure);
        qsort(sortedTable,size,sizeof(double),QsortComparator);

        //...	discarding
        //...	prevent before approx=0.(9) while error==1
        threshold=sortedTable[0];
        for(int i=size-1;i>=0;i--)
        {
            discardSum+=sortedTable[i]*sortedTable[i];
            approx=(sqrt(discardSum)/measure);
            if(approx>=error)
            {
                threshold=sortedTable[i];
                compress=static_cast<double>(i)*100.0/size;
                break;
            }
        }
        for(uint k=0;k<size;k++)
        {
            if(fabs(table[k])<threshold)
            {
                table[k]=0;
            }
        }
    }
    else
    {
    	compress=0.0;
    }
    delete []sortedTable;
}

//... CosineFFT implementation
void CosineFFT::CosFFT(double *y,const uint n,const Direction isign)
{
	uint i;
	double sum,sum1,y1,y2,ytemp;
	double theta,wi=0.0,wi1,wpi,wpr,wr=1.0,wr1,wtemp;

	theta=M_PI_2/n;
	wr1=cos(theta);
	wi1=sin(theta);
	wpr = -2.0*wi1*wi1;
	wpi=sin(2.0*theta);
	if (isign==Forward)
    {
		for (i=1;i<=n/2;i++)
        {
			y1=0.5*(y[i]+y[n-i+1]);
			y2=wi1*(y[i]-y[n-i+1]);
			y[i]=y1+y2;
			y[n-i+1]=y1-y2;
			wr1=(wtemp=wr1)*wpr-wi1*wpi+wr1;
			wi1=wi1*wpr+wtemp*wpi+wi1;
		}
		realFFT(y,n,isign);
		for (i=3;i<=n;i+=2)
        {
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
			y1=y[i]*wr-y[i+1]*wi;
			y2=y[i+1]*wr+y[i]*wi;
			y[i]=y1;
			y[i+1]=y2;
		}
		sum=0.5*y[2];
		for (i=n;i>=2;i-=2)
        {
			sum1=sum;
			sum += y[i];
			y[i]=sum1;
		}
	}
    else
    {
		ytemp=y[n];
		for (i=n;i>=4;i-=2)
        {
            y[i]=y[i-2]-y[i];
        }
		y[2]=2.0*ytemp;
		for (i=3;i<=n;i+=2)
        {
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
			y1=y[i]*wr+y[i+1]*wi;
			y2=y[i+1]*wr-y[i]*wi;
			y[i]=y1;
			y[i+1]=y2;
		}
		realFFT(y,n,isign);
		for (i=1;i<=n/2;i++)
        {
			y1=y[i]+y[n-i+1];
			y2=(0.5/wi1)*(y[i]-y[n-i+1]);
			y[i]=0.5*(y1+y2);
			y[n-i+1]=0.5*(y1-y2);
			wr1=(wtemp=wr1)*wpr-wi1*wpi+wr1;
			wi1=wi1*wpr+wtemp*wpi+wi1;
		}
        for(i=1;i<n+1;i++)
        {
            y[i]*=2.0/n;
        }
	}
}

//...	Daubechies4DWT implementation
void Daubechies4DWT::Daub4DWT(double *a,const unsigned int n,const Direction isign)
{
	double       *wksp;
	unsigned int nh,nh1,i,j;
	if(n<4)
	{
		return;
	}
	wksp=new double[n+1];
	nh1=(nh=n >> 1)+1;
	if(isign==Forward)
	{
		for(i=1,j=1;j<=n-3;j+=2,i++)
		{
			wksp[i]=C0*a[j]+C1*a[j+1]+C2*a[j+2]+C3*a[j+3];
			wksp[i+nh] = C3*a[j]-C2*a[j+1]+C1*a[j+2]-C0*a[j+3];
		}
		wksp[i]=C0*a[n-1]+C1*a[n]+C2*a[1]+C3*a[2];
		wksp[i+nh] = C3*a[n-1]-C2*a[n]+C1*a[1]-C0*a[2];
	}
	else
	{
		wksp[1]=C2*a[nh]+C1*a[n]+C0*a[1]+C3*a[nh1];
		wksp[2] = C3*a[nh]-C0*a[n]+C1*a[1]-C2*a[nh1];
		for(i=1,j=3;i<nh;i++)
		{
			wksp[j++]=C2*a[i]+C1*a[i+nh]+C0*a[i+1]+C3*a[i+nh1];
			wksp[j++] = C3*a[i]-C0*a[i+nh]+C1*a[i+1]-C2*a[i+nh1];
		}
	}
    memcpy(a+1,wksp+1,sizeof(double)*n);
	delete []wksp;
}


void HaarDWT::ForwardStep(const uint resolution)
{
	int level=resolution/2;
	for(int j=0;j<level;j++)
	{
		wavelet[j]=(ptr[j*2]+ptr[j*2+1])/M_SQRT2;
		wavelet[j+level]=(ptr[j*2]-ptr[j*2+1])/M_SQRT2;
	}
	for(uint k=0;k<resolution;k++)
	{
		ptr[k]=wavelet[k];
	}
}

void HaarDWT::_ForwardTransform(const uint size)
{
	int temp=size;
	while(temp>1)
	{
		ForwardStep(temp);
		temp/=2;
	}
}

void HaarDWT::InverseStep(const uint resolution)
{
	for(uint j=0;j<resolution;j+=2)
	{
		wavelet[j]=(ptr[j/2]+ptr[(j+resolution)/2])/M_SQRT2;
		wavelet[j+1]=(ptr[j/2]-ptr[(j+resolution)/2])/M_SQRT2;
	}
	for(uint k=0;k<resolution;k++)
	{
		ptr[k]=wavelet[k];
	}
}

void HaarDWT::_InverseTransform(const uint size)
{
	int level=size;
	int temp=2;

	while(temp<=level)
	{
		InverseStep(temp);
		temp*=2;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
void HaarDWT2D::ForwardTransform(double *d,const uint nn)
{
    wavelet=new double[nn];
    ptr=new double[nn];
	//... converts rows
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j]=d[i*nn+j];
        }
        _ForwardTransform(nn);
        for(uint j=0;j<nn;j++)
        {
            d[i*nn+j]=ptr[j];
        }
    }
	//... converts columns
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j]=d[j*nn+i];
        }
        _ForwardTransform(nn);
        for(uint j=0;j<nn;j++)
        {
            d[j*nn+i]=ptr[j];
        }
    }
    delete []ptr;
    delete []wavelet;
}

void HaarDWT2D::InverseTransform(double *d,const uint nn)
{
    wavelet=new double[nn];
    ptr=new double[nn];
	//... converts columns
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j]=d[j*nn+i];
        }
        _InverseTransform(nn);
        for(uint j=0;j<nn;j++)
        {
            d[j*nn+i]=ptr[j];
        }
    }
	//... converts rows
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j]=d[i*nn+j];
        }
        _InverseTransform(nn);
        for(uint j=0;j<nn;j++)
        {
            d[i*nn+j]=ptr[j];
        }
    }
    delete []ptr;
    delete []wavelet;
}

void Daubechies4DWT2D::ForwardTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
	//... converts rows
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[i*nn+j];
        }
        Daub4DWT(ptr,nn,Forward);
        for(uint j=0;j<nn;j++)
        {
            d[i*nn+j]=ptr[j+1];
        }
    }
	//... converts columns
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[j*nn+i];
        }
        Daub4DWT(ptr,nn,Forward);
        for(uint j=0;j<nn;j++)
        {
            d[j*nn+i]=ptr[j+1];
        }
    }
    delete []ptr;
}

void Daubechies4DWT2D::InverseTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
	//... converts columns
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[j*nn+i];
        }
        Daub4DWT(ptr,nn,Inverse);
        for(uint j=0;j<nn;j++)
        {
            d[j*nn+i]=ptr[j+1];
        }
    }
	//... converts rows
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[i*nn+j];
        }
        Daub4DWT(ptr,nn,Inverse);
        for(uint j=0;j<nn;j++)
        {
            d[i*nn+j]=ptr[j+1];
        }
    }
    delete []ptr;
}

void CosineFFT2D::ForwardTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
	//... converts rows
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[i*nn+j];
        }
        CosFFT(ptr,nn,Forward);
        for(uint j=0;j<nn;j++)
        {
            d[i*nn+j]=ptr[j+1];
        }
    }
	//... converts columns
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[j*nn+i];
        }
        CosFFT(ptr,nn,Forward);
        for(uint j=0;j<nn;j++)
        {
            d[j*nn+i]=ptr[j+1];
        }
    }
    delete []ptr;
}

void CosineFFT2D::InverseTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
	//... converts columns
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[j*nn+i];
        }
        CosFFT(ptr,nn,Inverse);
        for(uint j=0;j<nn;j++)
        {
            d[j*nn+i]=ptr[j+1];
        }
    }
	//... converts rows
    for(uint i=0;i<nn;i++)
    {
        for(uint j=0;j<nn;j++)
        {
            ptr[j+1]=d[i*nn+j];
        }
        CosFFT(ptr,nn,Inverse);
        for(uint j=0;j<nn;j++)
        {
            d[i*nn+j]=ptr[j+1];
        }
    }
    delete []ptr;
}


////////////////////////////////////////////////////////////////////////////////////////////
Transformer::Transformer(const int s,const int g,const double e,Transform *t)
			:size(s),
             grid(g),
             error(e),
             transform(t),
             conv(s,s)
{
	gridSize=size/grid;
    bitmap=new double[size*size];
    compress=new double[grid*grid];
    field=new BitField<double> *[grid];
    for(int i=0;i<grid;i++)
    {
        field[i]=new BitField<double>[grid];
        for(int j=0;j<grid;j++)
        {
            field[i][j].Init(gridSize,gridSize);
        }
    }
}

Transformer::~Transformer()
{
    delete []bitmap;
    delete []compress;
    for(int i=0;i<grid;i++)
    {
        delete []field[i];
    }
    delete []field;
}

double Transformer::Compress()
{
	int i;
    double compression=0;
    for(i=0;i<grid*grid;i++)
    {
       	compression+=compress[i];
    }
    return compression/i;
}

void Transformer::CopyToFields(const Zig zig)
{
	if(zig==ZigZag)
    {
        for(int i=0;i<grid;i++)
        {
            for(int j=0;j<grid;j++)
            {
            	ZigZagIterator<double> aIterator(field[i][j]);
                for(int k=0;k<gridSize;k++)
                {
                    for(int l=0;l<gridSize;l++)
                    {
                        aIterator.Current()=bitmap[conv(gridSize*j+l,gridSize*i+k)];
                        aIterator++;
                    }
                }
            }
        }
    }
    else
    {
        for(int i=0;i<grid;i++)
        {
            for(int j=0;j<grid;j++)
            {
            	BitFieldIterator<double> aIterator(field[i][j]);
                for(int k=0;k<gridSize;k++)
                {
                    for(int l=0;l<gridSize;l++)
                    {
                        aIterator.Current()=bitmap[conv(gridSize*j+l,gridSize*i+k)];
                        aIterator++;
                    }
                }
            }
        }
    }
 }

void Transformer::DoTransform()
{
    for(int i=0;i<grid;i++)
    {
        for(int j=0;j<grid;j++)
        {
            //... and here we transform it
            transform->ForwardTransform(field[i][j],gridSize);
        }
    }
}

void Transformer::CopyToTable(const Zig zig)
{
    if(zig==ZigZag)
    {
        for(int i=0;i<grid;i++)
        {
            for(int j=0;j<grid;j++)
            {
                //... and put back to the 'bitmap'
                ZigZagIterator<double> aIterator(field[i][j]);
                for(int k=0;k<gridSize;k++)
                {
                    for(int l=0;l<gridSize;l++)
                    {
                        bitmap[conv(gridSize*j+l,gridSize*i+k)]=aIterator.Current();
                        aIterator++;
                    }
                }
            }
        }
    }
    else
    {
		for(int i=0;i<grid;i++)
        {
            for(int j=0;j<grid;j++)
            {
                //... and put back to the 'bitmap'
                BitFieldIterator<double> aIterator(field[i][j]);
                for(int k=0;k<gridSize;k++)
                {
                    for(int l=0;l<gridSize;l++)
                    {
                        bitmap[conv(gridSize*j+l,gridSize*i+k)]=aIterator.Current();
                        aIterator++;
                    }
                }
            }
        }
    }
}

void Transformer::SortAndDiscard()
{
    for(int i=0;i<grid;i++)
    {
        for(int j=0;j<grid;j++)
        {
			transform->SortAndDiscard(field[i][j].field,gridSize,error);
            compress[i*grid+j]=transform->Compress();
        }
    }
}

void Transformer::DoInverseTransform()
{
    for(int i=0;i<grid;i++)
    {
        for(int j=0;j<grid;j++)
        {
            transform->InverseTransform(field[i][j],gridSize);
        }
    }
}

void Transformer::Process(const Zig zig)
{
    CopyToFields(zig);
    DoTransform();
    CopyToTable(Normal);
    SortAndDiscard();
    DoInverseTransform();
    CopyToTable(zig);
}

#ifdef MAIN

const int WIDTH=1024;
const int TABLE_SIZE=WIDTH*WIDTH;

int main()
{
    windowHandle
    return 0;
}

/*
    ofsteam ofs("DB4.dat");
    double wavelet[WIDTH];
    for(int i=0;i<WIDTH;i++)
    {
        wavelet[i]=0.0;
    }
    wavelet[256]=1.0;

    Daubechies4DWT transform(wavelet,WIDTH,Trabsform::Inverse);
    transform.InverseTransform(wavelet,WIDTH);
    transform.ReportTransform(wavelet,WIDTH,ofs)

	try
	{
    	Transformer transformer(WIDTH,2,0.1,new HaarDWT2D);
        for(int i=0;i<WIDTH;i++)
        {
	        for(int j=0;j<WIDTH;j++)
   	     	{
            	transformer[i*WIDTH+j]=(i*WIDTH+j);
        	}
        }
        transformer.Process(Transformer::Normal);
    }
	catch(xmsg &msg)
	{
		cout<<msg.why();
	}
*/
#endif //MAIN

