//---------------------------------------------------------------------------
#ifndef transformH
#define transformH
#include <values.h>
typedef unsigned int uint;

template <class T>
inline
void Swap(T &a,T &b)
{
    T tempr=a;
    a=b;
    b=tempr;
}

template <class T>
inline
T Max(T a,T b)
{
    return a>b ? a : b;
}

template <class T>
inline
T Min(T a,T b)
{
    return a<b ? a : b;
}

inline
ostream &tab(ostream &os)
{
    return os<<"\t";
}

template <class T,const int limit>
class RoundTo
{
	T round;
public:
	RoundTo(double d):round(0)	{	round=(d<limit ? (d>=0.0 ? d : 0) : limit);	}
    operator T()				{	return round;	}
};

template <class T,const int limit>
class ScaleTo
{
	double *table;
public:
	ScaleTo(double *,const int);
    void Init(double *,const int);
    double &operator [](const int index)				{	return table[index];	}
};

class Transform
{
public:
    enum Direction
    {
        Forward=1,
        Inverse=-1
    };
protected:
    double 	*ptr;
    double  compress;
    void 	realFFT(double *data,const unsigned int n,const Direction);
    void 	FFT(double *data,const unsigned int n,const Direction);
public:
	Transform():ptr(0),compress(0)		{	}
    double 			Compress() const	{	return compress;	}

    virtual void    ForwardTransform(double *,const uint)=0;
    virtual void    InverseTransform(double *,const uint)=0;
    virtual void  	SortAndDiscard(double *,const uint,const double);
    virtual void    ReportTransform(const double *,const uint,ofstream &);
};

class CosineFFT : public Transform
{
protected:
    void CosFFT(double *,const uint,const Direction);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class Daubechies4DWT : public Transform
{
protected:
    void Daub4DWT(double *,const uint,const Direction);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class HaarDWT : public Transform
{
protected:
    double *wavelet;

    virtual void ForwardStep(const uint);
    virtual void InverseStep(const uint);
    virtual void _ForwardTransform(const uint);
    virtual void _InverseTransform(const uint);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class HaarDWT2D : public HaarDWT
{
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class Daubechies4DWT2D : public Daubechies4DWT
{
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class CosineFFT2D : public CosineFFT
{
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

class Converter
{
    int 	height;
    int 	width;
public:
    Converter(const int w,const int h):width(w),height(h)   {   }

	//...	returns the position of X,Y point in 'width*height' bitmap;
    int 	operator  ()(const int w,const int h) 			{	return width*h+w;	}
};

template <class T>
class BitField
{
public:
	T 	*field;
    int X,Y;
public:
    BitField(const int x,const y):X(x),Y(y)		{	field=new T[X*Y];	}
    BitField():X(0),Y(0)						{	}
    ~BitField()									{	delete []field;	}
    void Init(const int x,const y)				{	field=new T[(X=x,Y=y),X*Y];	}

    T &operator ()(const int index)				{	return field[index];	}
    operator T *()								{	return field;	}
};

template <class T>
class BitFieldIterator
{
protected:
	BitField<T> 	&field;
    int         	cursor;
    int 			limit;
public:
	BitFieldIterator(BitField<T> &f):field(f),cursor(0),limit(field.X*field.Y)	{	}
    void   		operator ++(int);
	operator 	void *();
    T 			&Current()	 							{	return field[cursor];	}
    void 		Reset()	 								{	cursor=0;	}
};

template <class T>
BitFieldIterator<T>::operator void *()
{
	return cursor==limit ? 0 : this;
}

template <class T>
void BitFieldIterator<T>::operator ++(int)
{
	cursor++;
}

template <class T>
class ZigZagIterator : public BitFieldIterator<T>
{
	int zig;
	int ZigZag();
public:
	ZigZagIterator(BitField<T> &f):BitFieldIterator<T>(f),zig(0)	{	}
    void   		operator ++(int);
	operator 	void *();
    void 		Reset()	 											{	cursor=zig=0;	}
};

template <class T>
ZigZagIterator<T>::operator void *()
{
	return zig==limit ? 0 : this;
}

template <class T>
void ZigZagIterator<T>::operator ++(int)
{
	zig++;
	ZigZag();
}

template <class T>
int ZigZagIterator<T>::ZigZag()
{
    int index=0;
    int temp=0;

    int size=field.X;
    int x=zig/size;
    int y=zig%size;
    int diag=x+y;
    //...   x+y gives us the number of diagonal
    //      if it is odd the direction is from up to down
    if(diag<size)
    {
        for(int i=0;i<diag+1;i++)
        {
            temp+=i;
        }
        index+=((diag & 0x1) ? temp+y : temp+x);
    }
    else
    {
        temp=2*size-2;
        diag=temp-diag;
        temp=0;
        for(int i=0;i<diag+1;i++)
        {
            temp+=i;
        }
        index+=((diag & 0x1) ? temp+size-1-y: temp+size-1-x);
        index=size*size-1-index;
    }
    return cursor=index,cursor;
}

class Transformer
{
	double 				*bitmap;
    BitField<double>    **field;
	double     			*compress;
    double 				error;

	int					size;
    int					grid;
    int					gridSize;
    Converter 			conv;
	Transform     		*transform;
public:
	enum Zig
    {
    	ZigZag,
        Normal
    };
	Transformer(const int ,const int ,const double ,Transform * =0);
	~Transformer();
 	double &operator [](const int index)	{	return bitmap[index];	}
    double *Bitmap()						{	return bitmap;	}
    void Process(const Zig=ZigZag);
    	void CopyToFields(const Zig);
    	void CopyToTable(const Zig);
        void DoTransform();
		void SortAndDiscard();
        void DoInverseTransform();

    double Compress();
};

//... inlines
inline
void CosineFFT::ForwardTransform(double *d,const uint nn)
{
	int n=nn*nn;
    ptr=new double[n+1];
    memcpy(ptr+1,d,sizeof(double)*n);

    CosFFT(ptr,n,Forward);

    memcpy(d,ptr+1,sizeof(double)*n);
    delete []ptr;
}

inline
void CosineFFT::InverseTransform(double *d,const uint nn)
{
	int n=nn*nn;
    ptr=new double[n+1];
    memcpy(ptr+1,d,sizeof(double)*n);

    CosFFT(ptr,n,Inverse);

    memcpy(d,ptr+1,sizeof(double)*n);
    delete []ptr;
};

inline
void Daubechies4DWT::ForwardTransform(double *d,const uint nn)
{
	int n=nn*nn;
    ptr=new double[n+1];
    memcpy(ptr+1,d,sizeof(double)*n);

    Daub4DWT(ptr,n,Forward);

    memcpy(d,ptr+1,sizeof(double)*n);
    delete []ptr;
}

inline
void Daubechies4DWT::InverseTransform(double *d,const uint nn)
{
	int n=nn*nn;
    ptr=new double[n+1];
    memcpy(ptr+1,d,sizeof(double)*n);

    Daub4DWT(ptr,n,Inverse);

    memcpy(d,ptr+1,sizeof(double)*n);
    delete []ptr;
}

inline
void HaarDWT::ForwardTransform(double *d,const uint nn)
{
	int n=nn*nn;
    wavelet=new double[n];
    ptr=d;
    _ForwardTransform(n);
    delete []wavelet;
}

inline
void HaarDWT::InverseTransform(double *d,const uint nn)
{
	int n=nn*nn;
    wavelet=new double[n];
    ptr=d;
    _InverseTransform(n);
    delete []wavelet;
}

inline
int QsortComparator(const void *p1, const void *p2)
{
	return (*static_cast<const double *>(p1) < *static_cast<const double *>(p2)) - (*static_cast<const double *>(p1) > *static_cast<const double *>(p2));
}

template <class T,const int limit>
ScaleTo<T,limit>::ScaleTo(double *t,const int size)
{
	Init(t,size);
}

template <class T,const int limit>
void ScaleTo<T,limit>::Init(double *t,const int size)
{
	double max=-MINDOUBLE;
	double min=MAXDOUBLE;
    double value;
    table=t;
    for(int i=0;i<size;i++)
    {
    	value=table[i];
        max=(value>max ? max=value : max);
        min=(value<min ? min=value : min);
    }
    if(max>limit || min<0.0)
	{
        for(int i=0;i<size;i++)
        {
            value=(table[i]+::Max(min,0.0))/static_cast<double>(limit)*(max-min);
            table[i]=RoundTo<T,limit>(value);
        }
    }
}
#endif
