//---------------------------------------------------------------------------
//#include <vcl\vcl.h>
//#pragma hdrstop
//---------------------------------------------------------------------------
//USEFORM("transfor.cpp", Form1);
//USERES("tt.res");
//---------------------------------------------------------------------------
#include <except.h>
#include <cstring.h>
#include <math.h>
#include <stdlib.h>
#include <fstream.h>

typedef unsigned int uint;

const int TABLE_SIZE=65536;

const double C0=0.4829629131445341;
const double C1=0.8365163037378079;
const double C2=0.2241438680420134;
const double C3=-0.1294095225512604;

template <class T>
inline
void Swap(T &a,T &b)
{
    T tempr=a;
    a=b;
    b=tempr;
}

ostream &tab(ostream &os)
{
    return os<<"\t";
}

class Transform
{
public:
    enum Direction
    {
        Forward=1,
        Inverse=-1
    };
    virtual void    ForwardTransform(double *,const uint)=0;
    virtual void    InverseTransform(double *,const uint)=0;
    virtual double  SortAndDiscard(double *,const uint,const double);
    virtual void    ReportTransform(const double *,const uint,ofstream &);
protected:
    double *ptr;
    void realFFT(double *data,const unsigned int n,const Direction);
    void FFT(double *data,const unsigned int n,const Direction);
};
               
void Transform::FFT(double *data,const unsigned int nn,const Direction isign)
{
	unsigned int n,mmax,m,j,istep,i;
	double wtemp,wr,wpr,wpi,wi,theta;
	double tempr,tempi;

	n=nn << 1;
	j=1;
	for (i=1;i<n;i+=2)
    {
		if (j > i)
        {
			Swap(data[j],data[i]);
			Swap(data[j+1],data[i+1]);
		}
		m=n >> 1;
		while (m >= 2 && j > m)
        {
			j -= m;
			m >>= 1;
		}
		j += m;
	}
	mmax=2;
	while (n > mmax)
    {
		istep=mmax << 1;
		theta=isign*(2*M_PI/mmax);
		wtemp=sin(0.5*theta);
		wpr = -2.0*wtemp*wtemp;
		wpi=sin(theta);
		wr=1.0;
		wi=0.0;
		for (m=1;m<mmax;m+=2)
        {
			for (i=m;i<=n;i+=istep)
            {
				j=i+mmax;
				tempr=wr*data[j]-wi*data[j+1];
				tempi=wr*data[j+1]+wi*data[j];
				data[j]=data[i]-tempr;
				data[j+1]=data[i+1]-tempi;
				data[i] += tempr;
				data[i+1] += tempi;
			}
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
		}
		mmax=istep;
	}
}

void Transform::realFFT(double *data,const unsigned int n,const Direction isign)
{
	unsigned int i,i1,i2,i3,i4,np3;
	double c1=0.5,c2,h1r,h1i,h2r,h2i;
	double wr,wi,wpr,wpi,wtemp,theta;

	theta=M_PI/(double) (n>>1);
	if (isign==Forward)
    {
		c2 = -0.5;
		FFT(data,n>>1,isign);
	}
    else
    {
		c2=0.5;
		theta = -theta;
	}
	wtemp=sin(0.5*theta);
	wpr = -2.0*wtemp*wtemp;
	wpi=sin(theta);
	wr=1.0+wpr;
	wi=wpi;
	np3=n+3;
	for(i=2;i<=(n>>2);i++)
    {
		i4=1+(i3=np3-(i2=1+(i1=i+i-1)));
		h1r=c1*(data[i1]+data[i3]);
		h1i=c1*(data[i2]-data[i4]);
		h2r = -c2*(data[i2]+data[i4]);
		h2i=c2*(data[i1]-data[i3]);
		data[i1]=h1r+wr*h2r-wi*h2i;
		data[i2]=h1i+wr*h2i+wi*h2r;
		data[i3]=h1r-wr*h2r+wi*h2i;
		data[i4] = -h1i+wr*h2i+wi*h2r;
		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	if(isign==Forward)
    {
		data[1] = (h1r=data[1])+data[2];
		data[2] = h1r-data[2];
	}
    else
    {
		data[1]=c1*((h1r=data[1])+data[2]);
		data[2]=c1*(h1r-data[2]);
		FFT(data,n>>1,isign);
	}
}

void Transform::ReportTransform(const double *table,const uint nn,ofstream &os)
{
   	for(uint i=0;i<nn;i++)
    {
   		os<<table[i]<<tab;
    }
}

inline int QsortComparator(const void *p1, const void *p2)
{
	return (*reinterpret_cast<const double *>(p1) < *reinterpret_cast<const double *>(p2)) - (*reinterpret_cast<const double *>(p1) > *reinterpret_cast<const double *>(p2));
}

double Transform::SortAndDiscard(double *table,const uint size,const double error)
{
	double 	measure=0;
	double 	discardSum=0;

	double 	approx;
	double 	threshold;
	double  *sortedTable=new double[size];
    uint i;
    for(i=0;i<size;i++)
    {
        sortedTable[i]=fabs(table[i]);
        measure+=sortedTable[i]*sortedTable[i];
    }
    measure=sqrt(measure);
	qsort(sortedTable,size,sizeof(double),QsortComparator);

	//...	discarding
	//...	prevent before approx=0.(9) while error==1
	threshold=sortedTable[0];

    for(i=size-1;i>=0;i--)
    {
        discardSum+=sortedTable[i]*sortedTable[i];
        approx=(sqrt(discardSum)/measure);
        if(approx>=error)
        {
            threshold=sortedTable[i];
            break;
        }
    }
	for(uint k=0;k<size;k++)
	{
    	if(fabs(table[k])<threshold)
		{
			table[k]=0;
		}
	}
	delete []sortedTable;
	return double(i)/size*100.0;
}

class CosineFFT : public Transform
{
    void CosFFT(double *,const uint,const Direction);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

void CosineFFT::CosFFT(double *y,const uint n,const Direction isign)
{
	uint i;
	double sum,sum1,y1,y2,ytemp;
	double theta,wi=0.0,wi1,wpi,wpr,wr=1.0,wr1,wtemp;

	theta=M_PI_2/n;
	wr1=cos(theta);
	wi1=sin(theta);
	wpr = -2.0*wi1*wi1;
	wpi=sin(2.0*theta);
	if (isign==Forward)
    {
		for (i=1;i<=n/2;i++)
        {
			y1=0.5*(y[i]+y[n-i+1]);
			y2=wi1*(y[i]-y[n-i+1]);
			y[i]=y1+y2;
			y[n-i+1]=y1-y2;
			wr1=(wtemp=wr1)*wpr-wi1*wpi+wr1;
			wi1=wi1*wpr+wtemp*wpi+wi1;
		}
		realFFT(y,n,isign);
		for (i=3;i<=n;i+=2)
        {
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
			y1=y[i]*wr-y[i+1]*wi;
			y2=y[i+1]*wr+y[i]*wi;
			y[i]=y1;
			y[i+1]=y2;
		}
		sum=0.5*y[2];
		for (i=n;i>=2;i-=2)
        {
			sum1=sum;
			sum += y[i];
			y[i]=sum1;
		}
	}
    else
    {
		ytemp=y[n];
		for (i=n;i>=4;i-=2)
        {
            y[i]=y[i-2]-y[i];
        }
		y[2]=2.0*ytemp;
		for (i=3;i<=n;i+=2)
        {
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
			y1=y[i]*wr+y[i+1]*wi;
			y2=y[i+1]*wr-y[i]*wi;
			y[i]=y1;
			y[i+1]=y2;
		}
		realFFT(y,n,isign);
		for (i=1;i<=n/2;i++)
        {
			y1=y[i]+y[n-i+1];
			y2=(0.5/wi1)*(y[i]-y[n-i+1]);
			y[i]=0.5*(y1+y2);
			y[n-i+1]=0.5*(y1-y2);
			wr1=(wtemp=wr1)*wpr-wi1*wpi+wr1;
			wi1=wi1*wpr+wtemp*wpi+wi1;
		}
        for(i=1;i<n+1;i++)
        {
            y[i]*=2.0/n;
        }
	}
}

inline
void CosineFFT::ForwardTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    CosFFT(ptr,nn,Forward);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
}

inline
void CosineFFT::InverseTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    CosFFT(ptr,nn,Inverse);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
};

class Daubechies4DWT : public Transform
{
    void Daub4DWT(double *,const uint,const Direction);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

inline
void Daubechies4DWT::ForwardTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    Daub4DWT(ptr,nn,Forward);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
}

inline
void Daubechies4DWT::InverseTransform(double *d,const uint nn)
{
    ptr=new double[nn+1];
    memcpy(ptr+1,d,sizeof(double)*nn);

    Daub4DWT(ptr,nn,Inverse);

    memcpy(d,ptr+1,sizeof(double)*nn);
    delete []ptr;
}

void Daubechies4DWT::Daub4DWT(double *a,const unsigned int n,const Direction isign)
{
	double       *wksp;
	unsigned int nh,nh1,i,j;
	if(n<4)
	{
		return;
	}
	wksp=new double[n+1];
	nh1=(nh=n >> 1)+1;
	if(isign==Forward)
	{
		for(i=1,j=1;j<=n-3;j+=2,i++)
		{
			wksp[i]=C0*a[j]+C1*a[j+1]+C2*a[j+2]+C3*a[j+3];
			wksp[i+nh] = C3*a[j]-C2*a[j+1]+C1*a[j+2]-C0*a[j+3];
		}
		wksp[i]=C0*a[n-1]+C1*a[n]+C2*a[1]+C3*a[2];
		wksp[i+nh] = C3*a[n-1]-C2*a[n]+C1*a[1]-C0*a[2];
	}
	else
	{
		wksp[1]=C2*a[nh]+C1*a[n]+C0*a[1]+C3*a[nh1];
		wksp[2] = C3*a[nh]-C0*a[n]+C1*a[1]-C2*a[nh1];
		for(i=1,j=3;i<nh;i++)
		{
			wksp[j++]=C2*a[i]+C1*a[i+nh]+C0*a[i+1]+C3*a[i+nh1];
			wksp[j++] = C3*a[i]-C0*a[i+nh]+C1*a[i+1]-C2*a[i+nh1];
		}
	}
	for(i=1;i<=n;i++)
	{
		a[i]=wksp[i];
	}
	delete []wksp;
}

class HaarDWT : public Transform
{
    double *wavelet;

    void ForwardStep(const uint);
    void InverseStep(const uint);
    void _ForwardTransform(const uint);
    void _InverseTransform(const uint);
public:
    virtual void ForwardTransform(double *,const uint);
    virtual void InverseTransform(double *,const uint);
};

inline
void HaarDWT::ForwardTransform(double *d,const uint nn)
{
    wavelet=new double[nn];
    ptr=d;
    _ForwardTransform(nn);
    delete []wavelet;
}

inline
void HaarDWT::InverseTransform(double *d,const uint nn)
{
    wavelet=new double[nn];
    ptr=d;
    _InverseTransform(nn);
    delete []wavelet;
}

void HaarDWT::ForwardStep(const uint resolution)
{
	int level=resolution/2;
	for(int j=0;j<level;j++)
	{
		wavelet[j]=(ptr[j*2]+ptr[j*2+1])/M_SQRT2;
		wavelet[j+level]=(ptr[j*2]-ptr[j*2+1])/M_SQRT2;
	}
	for(uint k=0;k<resolution;k++)
	{
		ptr[k]=wavelet[k];
	}
}

void HaarDWT::_ForwardTransform(const uint size)
{
	int temp=size;
	while(temp>1)
	{
		ForwardStep(temp);
		temp/=2;
	}
}

void HaarDWT::InverseStep(const uint resolution)
{
	for(uint j=0;j<resolution;j+=2)
	{
		wavelet[j]=(ptr[j/2]+ptr[(j+resolution)/2])/M_SQRT2;
		wavelet[j+1]=(ptr[j/2]-ptr[(j+resolution)/2])/M_SQRT2;
	}
	for(uint k=0;k<resolution;k++)
	{
		ptr[k]=wavelet[k];
	}
}

void HaarDWT::_InverseTransform(const uint size)
{
	int level=size;
	int temp=2;

	while(temp<=level)
	{
		InverseStep(temp);
		temp*=2;
	}
}

int main()
{
	try
	{
        double tab[TABLE_SIZE];
        for(int i=0;i<TABLE_SIZE;i++)
        {
            tab[i]=i;
//            cout<<tab[i]<<endl;
        }
        CosineFFT cosFFT;
        cosFFT.ForwardTransform(tab,TABLE_SIZE);
        cout<<"\nTransform"<<endl;
        for(int i=0;i<TABLE_SIZE;i++)
        {
//            cout<<tab[i]<<endl;
        }
    /*
        cout<<"\nDiscarded coeffs : "<<(100-cosFFT.SortAndDiscard(tab,TABLE_SIZE,0.1))<<"%"<<endl;
        for(i=0;i<TABLE_SIZE;i++)
        {
            cout<<tab[i]<<endl;
        }
    */
        cosFFT.InverseTransform(tab,TABLE_SIZE);
        cout<<"\nInverse Transform"<<endl;
        for(int i=0;i<TABLE_SIZE;i++)
        {
//            cout<<tab[i]<<endl;
        }
        cout<<endl;

        cout<<"\ndaubFFT Transform"<<endl;
        Daubechies4DWT daubFFT;
        daubFFT.ForwardTransform(tab,TABLE_SIZE);
        for(int i=0;i<TABLE_SIZE;i++)
        {
//            cout<<tab[i]<<endl;
        }
        cout<<"\ndaubFFT Inverse Transform"<<endl;
        daubFFT.InverseTransform(tab,TABLE_SIZE);
        for(int i=0;i<TABLE_SIZE;i++)
        {
//            cout<<tab[i]<<endl;
        }
        cout<<endl;

        for(int i=0;i<TABLE_SIZE;i++)
        {
//            tab[i]=i;
        }

        cout<<"\nHaarFFT Transform"<<endl;
        HaarDWT haarFFT;
        haarFFT.ForwardTransform(tab,TABLE_SIZE);
        for(int i=0;i<TABLE_SIZE;i++)
        {
//            cout<<tab[i]<<endl;
        }
        cout<<"\nHaarFFT Inverse Transform"<<endl;
        haarFFT.InverseTransform(tab,TABLE_SIZE);
        for(int i=0;i<10;i++)
        {
            cout<<tab[i]<<endl;
        }
        cout<<endl;
        char t;
        cin>>t;
	}
	catch(xmsg &msg)
	{
		cout<<msg.why();
	}
	return 0;
}
//---------------------------------------------------------------------------
